#define __KERNEL__		/* kernel code */ 
#define MODULE			/* always as a module */ 

#include <linux/module.h>
#include <linux/version.h>

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>

#include <linux/fs.h>		/* for file_operations */
#include <linux/slab.h>      	/* for kmalloc() */
#include <asm/segment.h>	/* for memcpy */
#include <asm/io.h>		/* for virt_to_phys */

#include "debug.h"
#include "dmabuf.h"

EXPORT_NO_SYMBOLS; 

MODULE_AUTHOR("Guto");
MODULE_DESCRIPTION("DMA Buffer");

/* MODULE PARAMETERS */
int dev = DMABUF_MAJOR;
int n_bufs = DMABUF_N_DEVS;
int size = DMABUF_SIZE;
MODULE_PARM(dev,"i");
MODULE_PARM_DESC(dev,"DMA Buffer major device number");
MODULE_PARM(n_bufs,"i");
MODULE_PARM_DESC(n_bufs,"Number of DMA Buffer devices");
MODULE_PARM(size,"i");
MODULE_PARM_DESC(size,"DMA Buffer size (in bytes)");

/* PROTOTYPES */
int init_module(void); 
void cleanup_module(void);
static int dmabuf_open(struct inode *inode, struct file *filp);
static int dmabuf_close(struct file *filp);
static int dmabuf_mmap(struct file* filp, struct vm_area_struct* vma);
static int dmabuf_ioctl(struct inode* inode, struct file* filp, 
			unsigned int cmd, unsigned long arg);

/* DMA BUFFER FILE OPERATIONS */
static struct file_operations dmabuf_fops = {
    NULL, /* module */
    NULL, /* lseek */
    NULL, /* read */
    NULL, /* write */
    NULL, /* readdir */
    NULL, /* select */
    dmabuf_ioctl, /* ioctl */
    dmabuf_mmap, /* mmap */
    dmabuf_open, /* open */
    dmabuf_close, /* close */
    NULL, /* release */
    NULL, /* fsync */
    NULL, /* fasync */
    NULL, /* check_media_change */
    NULL, /* revalidate */
    NULL  /* lock */
};

/* DMA BUFFERS */
static struct DMA_Buffer *dmabufs;

/* DMA BUFFER SIZE AS POWER OF 2 TIMES PAGE_SIZE */
static int order;

int init_module(void)
{
    int i, err;

    debug(DB_TRACE, "Loading DMA Buffer(dev=%d,n_bufs=%d,size=%d)\n",
	  dev, n_bufs, size);

    /* Register the device driver */
    err = register_chrdev(dev, "dmabuf", &dmabuf_fops);
    if(err < 0)
	return err;

    /* Determine the DMA buffer size as 2^order pages */
    order = 0;
    for(i = size / PAGE_SIZE; i > 1; i >>= 1)
	order++; 
    debug(DB_INFO, "dmabuf->size=%d(2^%d*%ld)\n",	size, order, PAGE_SIZE);

    /* Allocate the control structure for DMA buffers */
    dmabufs = kmalloc(n_bufs * sizeof(struct DMA_Buffer), GFP_KERNEL);
    if(!dmabufs) {
	unregister_chrdev(dev, "dmabuf");
	return -ENOMEM ;
    }
    memset(dmabufs, 0, n_bufs * sizeof(struct DMA_Buffer));

    return 0;
}

void cleanup_module(void)
{
    debug(DB_TRACE, "Unloading DMA Buffer\n");

    /* Unregister the device driver */
    unregister_chrdev(dev, "dmabuf");

    /* Release DMA buffers control structure */
    kfree(dmabufs);
}

int dmabuf_open(struct inode *inode, struct file *filp)
{
    int i; 
    int unit;

    unit = MINOR(inode->i_rdev);

    debug(DB_TRACE, "dmabuf_open(unit=%d)\n", unit);

    if(unit >= n_bufs)
	return -ENODEV ;

    /* DMA buffers are not sharable */
    if(dmabufs[unit].in_use)
	return -EBUSY;

    /* Allocate the DMA buffer (2^order pages) */
    dmabufs[unit].buffer = (void *)__get_dma_pages(GFP_KERNEL, order);
    if(!dmabufs[unit].buffer)
	return -ENOMEM;
    dmabufs[unit].phy_addr = virt_to_phys(dmabufs[unit].buffer);

    /* Zero the DMA buffer */
    memset((void *) dmabufs[unit].buffer, 0, size);
    memcpy((void *) dmabufs[unit].buffer, "DMABUF", 6);

    /* Pin up the DMA buffer so it won't get swapped out */ 
    for (i = 0; i < size / PAGE_SIZE; i++) {
	set_bit(PG_reserved, &virt_to_page(dmabufs[unit].buffer + (i << PAGE_SHIFT))->flags);
	if(i)
	    atomic_inc(&virt_to_page(dmabufs[unit].buffer + (i << PAGE_SHIFT))->count);
    }
  
    debug(DB_INFO, "dmabufs[%d]->log_addr=%p\n", unit, dmabufs[unit].buffer);
    debug(DB_INFO, "dmabufs[%d]->phy_addr=%lx\n", unit, dmabufs[unit].phy_addr);
    debug(DB_INFO, "dmabufs[%d]->buffer=%s\n", unit, 
	  (char *)dmabufs[unit].buffer);

    dmabufs[unit].in_use = 1;
    MOD_INC_USE_COUNT;

    return 0;
}

int dmabuf_close(struct file *filp)
{
    int i; 
    int unit;

    unit = MINOR(filp->f_dentry->d_inode->i_rdev);

    debug(DB_TRACE, "dmabuf_close(unit=%d)\n", unit);

    debug(DB_INFO, "dmabufs[%d]->buffer[%p]=%s\n", unit, dmabufs[unit].buffer,
	  (char *)dmabufs[unit].buffer);

    /* Un-pin up the DMA buffer */
    for (i = 0; i < size / PAGE_SIZE; i++) {
	clear_bit(PG_reserved, &virt_to_page(dmabufs[unit].buffer + (i << PAGE_SHIFT))->flags);
	if(i)
	    atomic_dec(&virt_to_page(dmabufs[unit].buffer +
				     (i << PAGE_SHIFT))->count);
    }
 
    /* Release the DMA buffer */
    free_pages((unsigned long)dmabufs[unit].buffer, order);

    dmabufs[unit].in_use = 0;
    MOD_DEC_USE_COUNT;

    return 0;
}

int dmabuf_mmap(struct file* filp, struct vm_area_struct* vma)
{
    int unit;
    unsigned long vma_size, phy_addr;

    unit = MINOR(filp->f_dentry->d_inode->i_rdev);

    vma_size = vma->vm_end - vma->vm_start;

    debug(DB_TRACE, "dmabuf_mmap(unit=%d,start=%lx,size=%lx,off=%lx)\n",
	  unit, vma->vm_start, vma_size, vma->vm_pgoff * PAGE_SIZE);

    /* Mapping must be page aligned and not larger than the DMA buffer size */
    if((vma->vm_pgoff & (PAGE_SIZE - 1)) || (vma_size > size))
	return -EINVAL ;

    /* Get DMA buffer physical address for mapping */
    phy_addr = dmabufs[unit].phy_addr + vma->vm_pgoff * PAGE_SIZE;

    /* Map the DMA buffer in the requested address space range */
    if(remap_page_range( vma, vma->vm_start, phy_addr, vma_size, vma->vm_page_prot))
	return -EAGAIN ;

    filp->f_dentry->d_inode->i_count.counter++;

    return 0;
}

int dmabuf_ioctl(struct inode* inode, struct file* filp, unsigned int cmd,
		 unsigned long arg)
{
    int unit;

    unit = MINOR(filp->f_dentry->d_inode->i_rdev);

    debug(DB_TRACE, "dmabuf_ioctl(unit=%d,cmd=%x,arg=%lx)\n",
	  unit, cmd, arg);

    switch(cmd) {
    case DMABUF_GET_SIZE:  /* Return DMA buffer's size */
	*((int *)arg) = size;
	break;
    case DMABUF_GET_PHY_ADDR:  /* Return DMA buffer's physical address */
	*((unsigned long *)arg) = dmabufs[unit].phy_addr;
	break;
    default:
	return -EINVAL;
    }
    
    return 0;
}

