#include <string.h>
#include <debug.h>
#include <myrinet.h>

Myrinet::Myrinet(int unit)
{
    debug(DB_TRACE, "Myrinet(this=%p,unit=%d)\n", this, unit);

    mtu = 1024;
    dev = new PCI_Mapped_Device;
    if(!dev->is_valid()) {
	debug(DB_ERROR, "can't open /dev/dev0!\n");
	return;
    }

    if((dev->get_size() != sizeof(Myrinet_NIC))) {
	debug(DB_ERROR, "Wrong Myrinet size [%x]!\n", sizeof(Myrinet_NIC));
	return;
    }

    nic = (Myrinet_NIC *)dev->get_log_addr();
    
    if((sizeof(Myrinet_SRAM) != ntohl(nic->prom.lanai_sram_size))) {
	debug(DB_ERROR, "Wrong Myrinet SRAM size [%x]!\n", sizeof(Myrinet_SRAM));
	return;
    }

    nic->reset();
    
    debug(DB_INFO, "Physical address: %lx\n", dev->get_phy_addr());
    debug(DB_INFO, "FPGA [%p] =>\n", &nic->fpga);
    debug(DB_INFO, "  control: 0x%x\n", ntohl(nic->fpga[0]));
    debug(DB_INFO, "EEPROM =>\n");
    debug(DB_INFO, "  CPU version: 0x%x\n", ntohs(nic->prom.lanai_cpu_version));
    debug(DB_INFO, "  clock: 0x%x\n", ntohl(nic->prom.lanai_clockval));
    debug(DB_INFO, "  delay: 0x%x\n", ntohs(nic->prom.delay_line_value));
    debug(DB_INFO, "  SRAM size: 0x%x\n", ntohl(nic->prom.lanai_sram_size));
    debug(DB_INFO, "REGS [%p] =>\n", &nic->regs);
    debug(DB_INFO, "  ISR: 0x%x\n", ntohl(nic->regs.isr));
    debug(DB_INFO, "  RTC: %d\n", ntohl(nic->regs.rtc));

    dmabuf = new DMA_Buffer;
    if(!dmabuf->is_valid()) {
	debug(DB_ERROR, "can't get a DMA buffer!\n");
	delete dev;
    }
}

Myrinet::~Myrinet()
{
    debug(DB_TRACE, "~Myrinet()\n");
}

bool Myrinet::reset_ok()
{
    debug(DB_TRACE, "reset()\n");

    return true;
}

bool Myrinet::send_ok(void *buf, unsigned int *len)
{
    *len = align_length(*len);
    debug(DB_TRACE, "send(buf=%p,len=%d,rtc=%d)\n",
	  buf, *len, ntohl(nic->regs.rtc));

    if(!length_ok(*len)) {
	debug(DB_WARNING, "message is too long!\n");
	return false;
    }

    return true;
}

bool Myrinet::receive_ok(void *buf, unsigned int *len)
{
    debug(DB_TRACE, "receive(buf=%p,len=%d,rtc=%d)\n",
	  buf, *len, ntohl(nic->regs.rtc));

    *len = align_length(*len);
    if(!length_ok(*len)) {
	debug(DB_WARNING, "message is too long (%d)!\n", *len);
	return false;
  }

    return true;
}

void Myrinet::drop()
{
    char drop;

    while(!(nic->regs.isr & htonl(TAIL_INT_BIT))) {
	while(!(nic->regs.isr & htonl(BYTE_RDY_BIT)));
	drop = nic->regs.rb;
    }
}

int Myrinet::reset()
{
    if(!reset_ok())
	return -1;

    nic->reset();
    debug(DB_INFO, "RTC => %u\n", ntohl(nic->regs.rtc));

    return 0;
}

int Myrinet::send(void *buf, unsigned int len)
{
    if(!send_ok(buf, &len))
	return -1;
    
    memcpy(dmabuf->get_log_addr(), buf, len);

    nic->regs.dma_dir = htonl(1);
    nic->regs.ear = htonl(dmabuf->get_phy_addr());
    nic->regs.lar = htonl(dev->get_phy_addr() + MYRINET_SRAM_BASE);
    nic->regs.dma_ctr = htonl(len);
    while(!(nic->regs.isr & htonl(DMA_INT_BIT)));
    debug(DB_INFO, "DMA host->nic(ear=%x,lar=%x,ctr=%d)\n",
	  ntohl(nic->regs.ear), ntohl(nic->regs.lar), ntohl(nic->regs.dma_ctr));
  
    while(!(nic->regs.isr & htonl(SEND_RDY_BIT)));
    nic->regs.sh = htonl(MYRINET_PACKET_TYPE);

    while(!(nic->regs.isr & htonl(SEND_RDY_BIT)));
    nic->regs.sw = htonl(len);


    nic->regs.smp = 0;
    nic->regs.smlt = htonl(len);
    while(!(nic->regs.isr & htonl(SEND_INT_BIT)));
    debug(DB_INFO, "DMA nic->net(smp=%x,sml=%x)\n",
	  ntohl(nic->regs.smp), ntohl(nic->regs.sml));

    return 0;
}

int Myrinet::receive(void *buf, unsigned int *len)
{
    unsigned short header;
    do {
	while(!(nic->regs.isr & htonl(HALF_RDY_BIT)));
	header = ntohs(nic->regs.rh);
	if(header != MYRINET_PACKET_TYPE) {
	    drop();
	}
    } while(header != MYRINET_PACKET_TYPE);

    while(!(nic->regs.isr & htonl(WORD_RDY_BIT)));
    unsigned int l_len = ntohl(nic->regs.rw);


    if(!receive_ok(buf, len)) {
//    printf("1RTC => %u\n", ntohl(nic->regs.rtc));
	drop();
	return -1;
    }

    if((l_len > *len)) {
//    printf("2RTC => %u\n", ntohl(nic->regs.rtc));
	drop();
	return -1;
    }
    *len = l_len;

    nic->regs.rmp = 0x1000;
    nic->regs.rml = 0x1000 + mtu; //htonl(*len - sizeof(int));
//  while(!(nic->regs.isr & htonl(RECV_INT_BIT | BUFF_INT_BIT)));
    debug(DB_INFO, "DMA net->nic(rmp=%x,rml=%x)\n",
	  ntohl(nic->regs.rmp), ntohl(nic->regs.rml));

    while(!(nic->regs.isr & htonl(BYTE_RDY_BIT)));
    unsigned char crc = ntohl(nic->regs.rb);

/*    if(!(nic->regs.isr & htonl(TAIL_INT_BIT))) {
	drop();
	return -1;
    }
*/

//    memcpy(dmabuf->get_log_addr(), &nic->sram[0], *len);
    nic->regs.dma_dir = 0;
    nic->regs.ear = htonl(dmabuf->get_phy_addr());
    nic->regs.lar = htonl(dev->get_phy_addr() + MYRINET_SRAM_BASE);
    nic->regs.dma_ctr = htonl(*len);
    while(!(nic->regs.isr & htonl(DMA_INT_BIT)));
    debug(DB_INFO, "DMA host->nic(ear=%x,lar=%x,ctr=%d)\n",
	  ntohl(nic->regs.ear), ntohl(nic->regs.lar), ntohl(nic->regs.dma_ctr));


/*  printf("\nSRAM contents:\n");
  for(unsigned int i = 0; i < MYRINET_SRAM_SIZE; i++)
  {
    char value = nic->sram[i];
    if ((unsigned int) value != 0) printf("sram[%d] = %c\n", i, value);
  }
  printf("\n");


    printf("buf => %s\n", (char *) dmabuf->get_log_addr());
*/
    memcpy(buf, dmabuf->get_log_addr(), *len);

//    printf("buf => %s\n", (char *) buf);

    if(crc) {
//    printf("4RTC => %u\n", ntohl(nic->regs.rtc));
	debug(DB_WARNING, "CRC (%d)error!\n", crc);
	return -1;
    }

    return 0;
}
