#define __KERNEL__		/* kernel code */ 
#define MODULE			/* always as a module */ 

#include <linux/module.h>
#include <linux/version.h>

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>

#include <linux/fs.h>		/* for file_operations */
#include <linux/pci.h>      	/* for pci_find_device */
#include <linux/slab.h>      	/* for kmalloc */
#include <asm/segment.h>	/* for memcpy */
#include <asm/io.h>		/* for virt_to_phys */
#include <linux/byteorder/little_endian.h>

#include "debug.h"
#include "pcimap.h"

EXPORT_NO_SYMBOLS; 
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Guto");
MODULE_DESCRIPTION("PCI Device Mapper");

/* MODULE PARAMETERS */
int dev = PCIMAP_MAJOR;
int n_devs = PCIMAP_N_DEVS;
int vendor_id = -1;
int device_id = -1;
MODULE_PARM(dev,"i");
MODULE_PARM_DESC(dev,"PCI Map major device number");
MODULE_PARM(n_devs,"i");
MODULE_PARM_DESC(n_devs,"Number of PCI Map devices");
MODULE_PARM(vendor_id,"i");
MODULE_PARM_DESC(vendor_id,"PCI vendor id");
MODULE_PARM(device_id,"i");
MODULE_PARM_DESC(device_id,"PCI device id");

/* PROTOTYPES */
int init_module(void); 
void cleanup_module(void);
static int pcimap_open(struct inode *inode, struct file *filp);
static int pcimap_close(struct file *filp);
static int pcimap_mmap(struct file* filp, struct vm_area_struct* vma);
static int pcimap_ioctl(struct inode* inode, struct file* filp, 
			unsigned int cmd, unsigned long arg);

/* DMA BUFFER FILE OPERATIONS */
static struct file_operations pcimap_fops = {
    NULL, /* module */
    NULL, /* lseek */
    NULL, /* read */
    NULL, /* write */
    NULL, /* readdir */
    NULL, /* select */
    pcimap_ioctl, /* ioctl */
    pcimap_mmap, /* mmap */
    pcimap_open, /* open */
    pcimap_close, /* close */
    NULL, /* release */
    NULL, /* fsync */
    NULL, /* fasync */
    NULL, /* lock */
    NULL, /* readv */
    NULL, /* writev */
    NULL, /* sendpage */
    NULL  /* get_unmapped_aread */
};

/* Offset for PCI Regions */
static const int pciregs[PCIMAP_N_REGS] = {
    PCI_BASE_ADDRESS_0, 
    PCI_BASE_ADDRESS_1, 
    PCI_BASE_ADDRESS_2,
    PCI_BASE_ADDRESS_3, 
    PCI_BASE_ADDRESS_4,
    PCI_BASE_ADDRESS_5
};

/* Mapped PCI devices */
static struct PCI_Device *pcidevs;

int init_module(void)
{
    int i, j, err;
/*     void * log_addr; */
    struct pci_dev *pci_dev;
    
    debug(DB_TRACE,
	  "Loading PCI Map(dev=%d,n_devs=%d,vendor_id=%x,device_id=%x)\n",
	  dev, n_devs, vendor_id, device_id);

    /* Check for mandatory module parameters */
    if((vendor_id == -1) || (device_id == -1)) {
	printk(KERN_WARNING "PCI vendor and device id must be specified!\n");
	return -EINVAL;
    }

    /* Allocate the control structure for PCI Maps */
    pcidevs = kmalloc(n_devs * sizeof(struct PCI_Device), GFP_KERNEL);
    if(!pcidevs) 
	return -ENOMEM;
    memset(pcidevs, 0, n_devs * sizeof(struct PCI_Device));

    i = 0;
    pci_dev = NULL;
    do {
	/* Look for our PCI device */
	pci_dev = pci_find_device(vendor_id, device_id, pci_dev);
	if(!pci_dev)
	    break;

	/* (Re)configure device for bus mastering */
	pci_read_config_word(pci_dev, PCI_COMMAND, &pcidevs[i].cmd);
	if(!(pcidevs[i].cmd & PCI_COMMAND_MEMORY)) {
	    pcidevs[i].cmd |= PCI_COMMAND_MEMORY;
	    debug(DB_WARNING, "PCI_COMMAND_MEMORY not set for device!\n");
	}
	if(!(pcidevs[i].cmd & PCI_COMMAND_MASTER)) {
	    pcidevs[i].cmd |= PCI_COMMAND_MASTER;
	    debug(DB_WARNING, "PCI_COMMAND_MASTER not set for device!\n");
	}
	if(!(pcidevs[i].cmd & PCI_COMMAND_INVALIDATE)) {
	    pcidevs[i].cmd |= PCI_COMMAND_INVALIDATE;
	    debug(DB_WARNING, "PCI_COMMAND_INVALIDATE not set for device!\n");
	}
	pci_write_config_word(pci_dev, PCI_COMMAND, pcidevs[i].cmd);

	/* Get device's configuration */
	pci_read_config_word(pci_dev, PCI_COMMAND, &pcidevs[i].cmd);
	pci_read_config_word(pci_dev, PCI_STATUS, &pcidevs[i].stat);
	pci_read_config_byte(pci_dev, PCI_REVISION_ID, &pcidevs[i].rev);

	/* Get regions */
	for(j = 0; j < PCIMAP_N_REGS; j++) {

	    /* Get reagion info */
	    pci_read_config_dword(pci_dev, pciregs[j], &pcidevs[i].reg[j].phy_addr);
	    cli();
	    pci_write_config_dword(pci_dev, pciregs[j], ~0);
	    pci_read_config_dword(pci_dev, pciregs[j], &pcidevs[i].reg[j].size);
	    pci_write_config_dword(pci_dev, pciregs[j], pcidevs[i].reg[j].phy_addr);
	    sti();

	    /* Check for memory region 	*/
	    if(!pcidevs[i].reg[j].size 
	       || (pcidevs[i].reg[j].phy_addr & PCI_BASE_ADDRESS_SPACE_IO)) {
		pcidevs[i].reg[j].phy_addr = 0;
		pcidevs[i].reg[j].size = 0;
		continue;
	    }
 
	    /* We only handle 32 bit addressed memory above 1 Mb */
	    if((pcidevs[i].reg[j].phy_addr & PCI_BASE_ADDRESS_MEM_TYPE_MASK) &
	       (PCI_BASE_ADDRESS_MEM_TYPE_1M | PCI_BASE_ADDRESS_MEM_TYPE_64)) {
		pcidevs[i].reg[j].phy_addr = 0;
		pcidevs[i].reg[j].size = 0;
		continue;
	    }
	    
	    /* Adjust address and size */
	    pcidevs[i].reg[j].phy_addr &= PCI_BASE_ADDRESS_MEM_MASK;
	    pcidevs[i].reg[j].size =
		~(pcidevs[i].reg[j].size & PCI_BASE_ADDRESS_MEM_MASK) + 1;
	
	    pcidevs[i].regs |= 1 << j;
	}
	    
	if(pcidevs[i].regs)
	    pcidevs[i].found = 1;

	printk(KERN_WARNING "debug\n");
	debug(DB_INFO, "pcidev[%d]->ctrl=%x\n", i, pcidevs[i].cmd);
	debug(DB_INFO, "pcidev[%d]->stat=%x\n", i, pcidevs[i].stat);
	debug(DB_INFO, "pcidev[%d]->rev=%x\n", i, pcidevs[i].rev);
	debug(DB_INFO, "pcidev[%d]->regs=%x\n", i, pcidevs[i].regs);
	for(j = 0; j < PCIMAP_N_REGS; j++)
	    if(pcidevs[i].reg[j].size) {
		debug(DB_INFO, "pcidev[%d]->reg[%d]->phy_addr=%x\n", i, j,
		      pcidevs[i].reg[j].phy_addr);
		debug(DB_INFO, "pcidev[%d]->reg[%d]->size=%d\n", i, j,
		      pcidevs[i].reg[j].size);
	    }

    } while ((++i < n_devs) && pci_dev);

    if(i == 0){
	kfree(pcidevs);
	return -ENODEV;
    }

    SET_MODULE_OWNER (&pcimap_fops);

    /* Register the device driver */
    err = register_chrdev(dev, "pcimap", &pcimap_fops);
    if(err < 0)
	return err;
    
    return 0;
}

void cleanup_module(void)
{
    debug(DB_TRACE, "Unloading PCI Map\n");

    /* Unregister the device driver */
    unregister_chrdev(dev, "pcimap");

    /* Release PCI Maps control structure */
    kfree(pcidevs);
}

int pcimap_open(struct inode *inode, struct file *filp)
{
    int unit, reg;

    unit = MINOR(inode->i_rdev) >> 4;
    reg = MINOR(inode->i_rdev) & 0x0f;

    debug(DB_TRACE, "pcimap_open(unit=%d,reg=%d)\n", unit, reg);

    if((unit >= n_devs) || (!pcidevs[unit].found) || reg >= PCIMAP_N_REGS)
	return -ENODEV;

    if(!pcidevs[unit].reg[reg].size)
	return -ENODEV;

    /* PCI Map devices are not sharable */
    if(pcidevs[unit].reg[reg].busy)
	return -EBUSY;

    pcidevs[unit].reg[reg].busy = 1;

    return 0;
}

int pcimap_close(struct file *filp)
{
    int unit, reg;

    unit = MINOR(filp->f_dentry->d_inode->i_rdev) >> 4;
    reg = MINOR(filp->f_dentry->d_inode->i_rdev) & 0x0f;

    debug(DB_TRACE, "pcimap_close(unit=%d,reg=%d)\n", unit, reg);

    pcidevs[unit].reg[reg].busy = 0;

    return 0;
}

int pcimap_mmap(struct file* filp, struct vm_area_struct* vma)
{
    int unit, reg;
    unsigned long size, phy_addr;

    unit = MINOR(filp->f_dentry->d_inode->i_rdev) >> 4;
    reg = MINOR(filp->f_dentry->d_inode->i_rdev) & 0x0f;
 
    size = vma->vm_end - vma->vm_start;

    debug(DB_TRACE, "pcimap_mmap(unit=%d,reg=%d,start=%lx,size=%lx,off=%lx)\n",
	  unit, reg, vma->vm_start, size, vma->vm_pgoff * PAGE_SIZE);

    /* Mapping must be page aligned and not larger than the region size */
    if((size + vma->vm_pgoff * PAGE_SIZE) > pcidevs[unit].reg[reg].size)
	return -EINVAL ;

    /* Get device's memory physical address for mapping */
    phy_addr = pcidevs[unit].reg[reg].phy_addr + vma->vm_pgoff * PAGE_SIZE;

    /* It seems all the PCI address space is already mapped in the kernel */
    /*
    vma->vm_start = (unsigned int) ioremap_nocache(phy_addr, vma_size); 
    */

    /* Map device's memory in the requested address space range */
    if(remap_page_range(vma, vma->vm_start, phy_addr, size, vma->vm_page_prot))
	return -EAGAIN ;

    filp->f_dentry->d_inode->i_count.counter++;

    return 0;
}

int pcimap_ioctl(struct inode* inode, struct file* filp, unsigned int cmd,
		 unsigned long arg)
{
    int unit, reg;

    unit = MINOR(filp->f_dentry->d_inode->i_rdev) >> 4;
    reg = MINOR(filp->f_dentry->d_inode->i_rdev) & 0x0f;

    debug(DB_TRACE, "pcimap_ioctl(unit=%d,reg=%d,cmd=%x,arg=%lx)\n",
	  unit, reg, cmd, arg);

    switch(cmd) {
    case PCIMAP_GET_SIZE:  /* Return DMA buffer's size */
	*((int *)arg) = pcidevs[unit].reg[reg].size;
	break;
    case PCIMAP_GET_PHY_ADDR:  /* Return DMA buffer's physical address */
	*((unsigned long *)arg) = pcidevs[unit].reg[reg].phy_addr;
	break;
    default:
	return -EINVAL;
    }
    
    return 0;
}
